{$B-}

unit mn_polje; {mnozica, realizirana s poljem}

interface

uses mn_abstr;

const
  maxPolje=22; {najvecje stevilo elementov}

type
  kMnPolje=^mnPolje;
  mnPolje=object (mnozica) {glej mn_abstr}
    polje: array [0..maxPolje] of kVsebina; {polje, v katerem so elementi}
    zadnji: integer; {indeks zadnjega elementa}
    destructor mnDest; virtual;
    procedure naredi; virtual;
    procedure izvoziU (mn: kMnozica; preveri, iter: boolean); virtual;
    procedure izvoziP (mn, drugaMn: kMnozica; iter: boolean); virtual;
    function dodaj (vs: kVsebina; preveri, iter: boolean): boolean; virtual;
    function brisi (vs: kVsebina): boolean; virtual;
    function pripada (vs: kVsebina; iter: boolean): boolean; virtual;
    function unija (mn1, mn2: kMnozica; iter: boolean): boolean; virtual;
    function presek (mn1, mn2: kMnozica; iter: boolean): boolean; virtual;
  end;

implementation

destructor mnPolje.mnDest;
  var n: integer;
begin
  for n:=0 to zadnji do dispose (polje [n], vsDest); {sprosti pomnilnik, ki so ga zasedale vsebine}
end;

procedure mnPolje.naredi;
begin
  tip:=poljeEl;
  zadnji:=-1; {doloci, da je konec seznama pred zacetkom}
end;

procedure mnPolje.izvoziU (mn: kMnozica; preveri, iter: boolean);
  var
    n: integer;
    zac: kVsebina;
  procedure rekurzija (polozaj: integer); {procedura za rekurzijo - vsebina podobna kot pri iteraciji}
  begin
    if polozaj<=zadnji then begin
      zac:=polje [polozaj]^.prepisi;
      if not mn^.dodaj (zac, preveri, false) then dispose (zac, vsDest);
      rekurzija (polozaj+1);
    end;
  end;
begin
  if iter then
    for n:=0 to zadnji do begin {iterativno ...}
      zac:=polje [n]^.prepisi; {v zac se zapise kopija trenutnega elementa}
      if not mn^.dodaj (zac, preveri, true) then dispose (zac, vsDest); {ce dodajanje ne uspe, se zbrise}
    end
  else rekurzija (0); {... ali rekurzivno}
end;

procedure mnPolje.izvoziP (mn, drugaMn: kMnozica; iter: boolean);
  var n: integer;
  procedure rekurzija (polozaj: integer); {procedura za rekurzijo - vsebina podobna kot pri iteraciji}
  begin
    if polozaj<=zadnji then begin
      if drugaMn^.pripada (polje [polozaj], false) then
        mn^.dodaj (polje [polozaj]^.prepisi, false, false);
      rekurzija (polozaj+1);
    end;
  end;
begin
  if iter then
    for n:=0 to zadnji do begin {iterativno ...}
      if drugaMn^.pripada (polje [n], true) then     {ce trenutni element pripada drugaMn, ->}
        mn^.dodaj (polje [n]^.prepisi, false, true); {se njegova kopija zapise v mn}
    end
  else rekurzija (0); {... ali rekurzivno}
end;

function mnPolje.dodaj (vs: kVsebina; preveri, iter: boolean): boolean;
begin
  dodaj:=false;
  if (zadnji<maxPolje) and ((not preveri) or (not pripada (vs, iter))) then begin {ce je prostor in element ni ze vsebovan ->}
    inc (zadnji);                                                                 {ali pa tega ni treba preveriti, se doda ->}
    polje [zadnji]:=vs;                                                           {na konec}
    dodaj:=true;
  end;
end;

function mnPolje.brisi (vs: kVsebina): boolean;
  var n: integer;
begin
  brisi:=false;
  for n:=0 to zadnji do
    if vs^.jeEnak (polje [n]) then begin
      dispose (polje [n], vsDest); {sprosti spomin, ki ge je zasedal iskani element}
      polje [n]:=polje [zadnji]; {na njegovo mesto da zadnjega}
      dec (zadnji);
      brisi:=true;
      break;
    end;
end;

function mnPolje.pripada (vs: kVsebina; iter: boolean): boolean;
  var n: integer;
  procedure rekurzija (polozaj: integer); {procedura za rekurzijo - vsebina podobna kot pri iteraciji}
  begin
    if polozaj<=zadnji then
      if vs^.jeEnak (polje [polozaj]) then
        pripada:=true else rekurzija (polozaj+1);
  end;
begin
  pripada:=false;
  if iter then
    for n:=0 to zadnji do begin {iterativno ...}
      if vs^.jeEnak (polje [n]) then begin {ko prispe do iskanega elementa, zakljuci in vrne true}
        pripada:=true;
        break;
      end;
    end
  else rekurzija (0); {... ali rekurzivno}
end;

function mnPolje.unija (mn1, mn2: kMnozica; iter: boolean): boolean;
  var preveri: boolean;
begin
  if (mn1<>nil) and (mn2<>nil) then begin
    if (mn1<>@self) and (mn2<>@self) then begin {ce ni nobena izmed mnozic unije mnozica, v katero se unija zapisuje, se ->}
      mnDest;                                   {slednja izprazni, pri dodajanju mn1 pa ni treba preverjati pripadnosti}
      naredi;
      preveri:=false;
    end else preveri:=true;
    if (mn1<>@self) then mn1^.izvoziU (@self, preveri, iter); {obe mnozici se dodata v unijo, ce ni katera izmed njiju ze ->}
    if (mn2<>@self) then mn2^.izvoziU (@self, true, iter);    {mnozica, v ketero se unija zapisuje - uporabita se njuni metodi}
    unija:=true;
  end else unija:=false;
end;

function mnPolje.presek (mn1, mn2: kMnozica; iter: boolean): boolean;
  var
    n, popravek: integer;
    mn: kMnozica;
  procedure rekurzija (polozaj: integer); {procedura za rekurzijo - vsebina podobna kot pri iteraciji}
  begin
    if polozaj<=zadnji then begin
      if not mn^.pripada (polje [polozaj], false) then begin
        dispose (polje [polozaj], vsDest);
        polje [polozaj]:=polje [zadnji];
        dec (zadnji);
        rekurzija (polozaj);
      end else rekurzija (polozaj+1);
    end;
  end;
begin
  if (mn1<>nil) and (mn2<>nil) then begin
    if (mn1=@self) xor (mn2=@self) then begin {ce je ena izmed mnozic preseka mnozica, v katero se presek zapisuje}
      if (mn1=@self) then mn:=mn2 else mn:=mn1; {mn postane mnozica, v katero se ne zapisuje presek}
      if iter then begin
        popravek:=0;
        for n:=0 to zadnji do {kar ne spada v presek, se zbrise iterativno (na enak nacin kot v brisi) ...}
          if not mn^.pripada (polje [n-popravek], iter) then begin
            dispose (polje [n-popravek], vsDest);
            polje [n-popravek]:=polje [zadnji];
            dec (zadnji);
            inc (popravek);
          end;
      end else rekurzija (0); {... ali rekurzivno}
    end;
    if (mn1<>@self) and (mn2<>@self) then begin {ce ni nobena izmed mnozic preseka mnozica, v katero se presek zapisuje, se ->}
      mnDest;                                   {izprazni}
      naredi;
      mn1^.izvoziP (@self, mn2, iter); {dodajo se elementi mn1, ki ne pripadajo mn2 - uporabi se metoda mn1}
    end;
    presek:=true;
  end else presek:=false;
end;

end.